from io import StringIO
from html.parser import HTMLParser
import json
import requests
import urllib.parse
import urllib.request


def get_token(session):
    if not hasattr(session, "token") or not session.token:
        session.token = session.login()
    return session.token


def get_kinds(param):
    if "fileFormat" not in param.attributes:
        return []  # Default to an empty list

    # Get the list from the attributes
    kinds_string = param.attributes["fileFormat"]

    if kinds_string == "" or kinds_string == "*":
        return []  # Accepts all known kinds

    return kinds_string.split(";")


def get_permissions(job):
    url = f"{job.server_data.url}/rest/v1/jobs/{job.job_number}/permissions"
    request = urllib.request.Request(url)
    if job.server_data.authorization_header() is not None:
        request.add_header("Authorization", job.server_data.authorization_header())
    request.add_header("User-Agent", "GalaxyJupyterLabRest")

    response = urllib.request.urlopen(request)
    return json.loads(response.read())


def set_permissions(job, permissions):
    url = f"{job.server_data.url}/rest/v1/jobs/{job.job_number}/permissions"
    data = json.dumps(permissions).encode("utf8")
    request = urllib.request.Request(url, data=data, method="PUT")
    if job.server_data.authorization_header() is not None:
        request.add_header("Authorization", job.server_data.authorization_header())
    request.add_header("User-Agent", "GalaxyJupyterLabRest")
    urllib.request.urlopen(request)

    # def login(session):
    #    safe_url = ensure_safe_url(session._notebook_url)
    #    safe_email = urllib.parse.quote(session._notebook_email)
    #    safe_password = urllib.parse.quote(session._notebook_password)

    #    url = f"{safe_url}/rest/v1/oauth2/token?grant_type=password&username={safe_username}&password={safe_password}&client_id=GenePatternNotebook-{safe_username}"
    #    response = requests.post(url)

    #    try:
    #        response.raise_for_status()
    #        return response.json()['access_token']
    #    except requests.exceptions.Timeout:
    #        raise TimeoutError('Connection timed out attempting to contact GenePattern server')
    #    except requests.exceptions.TooManyRedirects:
    #        raise TimeoutError('Bad GenePattern server URL')
    #    except requests.exceptions.RequestException as e:
    #        raise TimeoutError('Invalid email or password')

    # def system_message(session):
    #    safe_url = ensure_safe_url(session._notebook_url)
    #    url = f"{safe_url}/rest/v1/config/system-message"
    #    response = requests.get(url)
    #    return strip_html(response.text)

    # def get_task(session, lsid_or_name): #ToDo
    #    url = f'{session._notebook_url}/rest/v1/tasks/{lsid_or_name}/'
    #    request = urllib.request.Request(url)
    #    if session.authorization_header() is not None:
    #        request.add_header('Authorization', session.authorization_header())
    #    request.add_header('User-Agent', 'GalaxyJupyterLabRest')#

    #    response = urllib.request.urlopen(request)
    #    task_dict = json.loads(response.read())
    #   return GalaxyTask(session, task_dict['lsid'], task_dict)

    # class HTMLStripper(HTMLParser):
    #    """Parse HTML blob and strip out all tags"""
    #    def __init__(self):
    #        super().__init__()
    #        self.reset()
    #        self.strict = False
    #        self.convert_charrefs = True
    #        self.text = StringIO()

    #    def error(self, message):
    #        pass

    #    def handle_data(self, d):
    #        self.text.write(d)

    #    def get_data(self):
    #        return self.text.getvalue()

    # def strip_html(html):
    """Strip potentially unsafe HTML from the system message string"""
    #    s = HTMLStripper()
    #    s.feed(html)
    #    return s.get_data()

    # def ensure_safe_url(url):
    """Ensure the Galaxy URL ends with /api"""


#    if url.endswith('/'):
#        url = url[:-1]
#    if not url.endswith('/api'):
#        url += '/api'
#    return url



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jayadev Joshi.
# Distributed under the terms of the Modified BSD License.

from nbtools import UIOutput

from ipywidgets import VBox, widget_serialization
from traitlets import Unicode, List, Dict, Instance
from ._frontend import module_name, module_version

# from .basewidget import BaseWidget


class GalaxyUIOutput(UIOutput):
    """
    Widget used to render Python output in a UI
    """

    _model_name = Unicode("GalaxyUIOutputModel").tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)

    _view_name = Unicode("GalaxyUIOutputView").tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    name = Unicode("Python Results").tag(sync=True)
    status = Unicode("").tag(sync=True)
    files = List(Unicode, []).tag(sync=True)
    text = Unicode("").tag(sync=True)
    visualization = Unicode("").tag(sync=True)
    appendix = Instance(VBox).tag(sync=True, **widget_serialization)
    extra_file_menu_items = Dict().tag(sync=True)

    def __init__(self, **kwargs):
        # Initialize the child widget container
        self.appendix = VBox()

        UIOutput.__init__(self, **kwargs)



import os
import IPython
from nbtools import NBTool
from .Galaxyuibuilder import GalaxyUIBuilder
from .util import DEFAULT_COLOR, DEFAULT_LOGO
import json5
import logging
import IPython.display
import glob
from nbtools import UIBuilder

import os
import re
import json


log = logging.getLogger(__name__)

try:
    from genepattern import authwidget
except:
    print("need to be fixed...!")
    pass

try:
    from gp import GPTask
except ModuleNotFoundError:
    log.warning(
        "gp module is not available, will not be able to mix GiN with GenePattern"
    )
    GPTask = None
import GiN

class GalaxyTaskWidget(GalaxyUIBuilder):
    """A widget for representing the status of a Galaxy job"""

    # default_color = 'rgba(10, 45, 105, 0.80)'
    
    default_color = DEFAULT_COLOR
    default_logo = DEFAULT_LOGO
    tool = None

    def __init__(self, tool=None, **kwargs):
       
        self.tool = tool

        a = GiN.sessions.SessionList()
        gi = a.get(server=self.tool['origin'])

        history_ids = gi.gi.histories.get_histories()

        if self.tool['name'] == 'workflow_explorer' or self.tool['name'] == 'Upload Data' :
            inputs = {'inpus':[]}
            history_data=[]
        else:
            inputs = gi.gi.tools.build(tool_id=self.tool["id"], history_id=history_ids[0]["id"])
            history_data = GalaxyTaskWidget.updated_form(
                server=self.tool['origin'],
                history_id=history_ids[0]["id"],
                python_side=True,
            )
        
        GalaxyUIBuilder.__init__(
            self,
            inputs=inputs,
            id = self.tool['id'],
            name = self.tool['name']+' ('+self.tool['origin']+')',
            galaxy_tool_id=self.tool['id'],
            description=self.tool['description'],
            subtitle=self.tool['origin'],
            history_ids=history_ids,
            history_data=history_data,
            color= self.default_color,
            logo=self.default_logo,
            origin=self.tool['origin'],
            **kwargs
        )

    def handle_error_task(self, error_message, name="Galaxy Module", **kwargs):
        """Display an error message if the task is None"""
        UIBuilder.__init__(self, lambda: None, color=self.default_color, **kwargs)

        self.name = name
        self.display_header = False
        self.display_footer = False
        self.error = error_message

    @staticmethod
    def RetrivParm(inputs, prefix=""):

        OutDict = {}

        for i in inputs:
            full_name = prefix + i["name"]

            OutDict[full_name] = i
            if i["model_class"] == "Conditional":
                for j in i["cases"]:
                    Dict1 = GalaxyTaskWidget.RetrivParm(j["inputs"], full_name + "|")
                    OutDict = dict(list(OutDict.items()) + list(Dict1.items()))
            elif i["model_class"] == "Repeat":
                Dict2 = GalaxyTaskWidget.RetrivParm(i["inputs"], full_name + "|")
                OutDict = dict(list(OutDict.items()) + list(Dict2.items()))
            elif i["model_class"] == "Section":
                Dict3 = GalaxyTaskWidget.RetrivParm(i["inputs"], full_name + "|")
                OutDict = dict(list(OutDict.items()) + list(Dict3.items()))

        return OutDict

    @staticmethod
    def submit_job( server, tool_id, tool_inputs=None, history_id=None):

        a = GiN.sessions.SessionList()
        gi1 = a.get(server=server)
        # tool_inputs = json5.loads(tool_inputs)
        new_inputs = GalaxyTaskWidget.RefinedInputs(tool_inputs, gi1)

        try:
            job = gi1.tools.gi.tools.run_tool(
                history_id=history_id, tool_id=tool_id, tool_inputs=new_inputs
            )
        except BaseException as e:
            job = {"state": "job failed", 'error': str(e)}


        #return IPython.display.JSON(job)
        return job

    @staticmethod
    def form_value( raw_value):

        """Give the default parameter value in format the UI Builder expects"""
        if raw_value is not None:
            return raw_value
        else:
            return ""

    @staticmethod
    def show_job( server, job_id):

        a = GiN.sessions.SessionList()
        gi1 = a.get(server=server)
        job = gi1.jobs.gi.jobs.show_job(job_id=job_id)

        # return IPython.display.JSON(job)
        return job

    
    @staticmethod
    def get_data_type_and_genomes( server=None):

        a = GiN.sessions.SessionList()
        gi2 = a.get(server=server)

        data_types = gi2.gi.datatypes.get_datatypes()
        genomes = gi2.gi.genomes.get_genomes()
        datatypes_genomes = {
            "datatypes": data_types,
            "genomes": genomes,

        }

        return IPython.display.JSON(datatypes_genomes)

    @staticmethod
    def Return_api_key( server):
        a = GiN.sessions.SessionList()
        gi = a.get(server=server)

        key = gi.gi.key
        email = gi.gi.users.get_current_user()['email']
        
        #return IPython.display.JSON({'api_key': key, 'email': email})
        return {'api_key': key, 'email': email}
    
    @staticmethod
    def upload_dataset(
         file_path, upload_method, datatype, genome, server=None, HistoryID=None
          
    ):  

        a = GiN.sessions.SessionList()
        gi3 = a.get(server=server)

        history = gi3.gi.histories.show_history(history_id=HistoryID)
        # a = hi.History(history, gi=gi3.gi)

        if upload_method == "text":
            job = gi3.gi.tools.put_url(content=file_path, history_id=HistoryID)
            del gi3
            return IPython.display.JSON(job)

        elif upload_method == "textarea":
            job = gi3.gi.tools.put_url(content=file_path, history_id=HistoryID)
            del gi3
            return IPython.display.JSON(job)

        # return IPython.display.JSON(job)
    @staticmethod
    def TestOut( server=None, JobID=None):

        a = GiN.sessions.SessionList()
        gi4 = a.get(server=server)

        status = gi4.jobs.gi.jobs.show_job(JobID, full_details=True)
       
        return IPython.display.JSON(status)

    @staticmethod
    def ReturnSessions( server=None):

        a = GiN.sessions.SessionList()
        gi4 = a.get(server=server)


    @staticmethod
    def OutPutData( server=None, JobID=None):

        a = GiN.sessions.SessionList()
        gi5 = a.get(server=server)

        showJob = gi5.jobs.gi.jobs.show_job(JobID, full_details=True)

        DataList = []

        for i in showJob["outputs"].keys():
            DataList.append(
                gi5.gi.datasets.gi.datasets.show_dataset(
                    dataset_id=showJob["outputs"][i]["id"],
                    hda_ldda=showJob["outputs"][i]["src"],
                )
            )

        return IPython.display.JSON(DataList)

    @staticmethod
    def RefinedInputs( inputs, gi):

        for i in inputs.keys():
            if type(inputs[i]) == dict:
                if list(inputs[i].keys())[0] == "values":
                    new_values = []
                    for j in inputs[i]["values"]:
                        new_values.append(j)
                    inputs[i]["values"] = new_values
        return inputs


    @staticmethod
    def return_session_list():
        try:
            a = GiN.sessions.SessionList()
            return a.get_servers()
        except:
             Python.display.JSON([])


    @staticmethod
    def updated_form(
        server=None,
        tool_inputs=None,
        tool_id=None,
        history_id=None,
        python_side=False,
        input_data_param=False,
    ):

        a = GiN.sessions.SessionList()
        gi6 = a.get(server=server)

        if (tool_inputs) and (tool_id):

            inputs = gi6.tools.gi.tools.build(
                tool_id=tool_id, inputs=tool_inputs, history_id=history_id
            )

            if input_data_param is False:
                # return IPython.display.JSON(data=inputs)
                data=inputs
                return data
            else:
                return GalaxyTaskWidget.RetrivParm(inputs["inputs"])
                    
        elif (tool_inputs is None) and (tool_id is not None):

            inputs = gi6.tools.gi.tools.build(tool_id=tool_id, history_id=history_id)
            data=inputs
            return data

        else:
            history_data = gi6.gi.datasets.gi.datasets.get_datasets(
                history_id=history_id,
                state="ok",
                deleted=False,
                purged=False,
                visible=True,
            )

            if python_side is True:
                return history_data
            else:
                # IPython.display.JSON(data=history_data)
                data=history_data
                return data


    @staticmethod
    def history_data_list( server=None, history_id=None):

        try:
            a = GiN.sessions.SessionList()
            gi7 = a.get(server=server)

            history_data = gi7.gi.datasets.gi.datasets.get_datasets(
                history_id=history_id, deleted=False, purged=False, visible=True
            )

            # # for i in HistoryData:
            # #  
            # #    datasets.append(gi.gi.datasets.gi.datasets.show_dataset(dataset_id=i['id']))
            #return IPython.display.JSON(history_data)
            return history_data

        except:

            #return IPython.display.JSON([])
            return []

    @staticmethod
    def show_data_set(server=None, dataset_id=None):

        a = GiN.sessions.SessionList()
        gi8 = a.get(server=server)
        show_dataset = gi8.gi.datasets.gi.datasets.show_dataset(dataset_id=dataset_id)
        
        #return IPython.display.JSON(show_dataset) #keeps kernel busy 
        return show_dataset

    @staticmethod
    def delete_dataset(server=None, history_id=None, dataset_id=None):

        a = GiN.sessions.SessionList()
        gi9 = a.get(server=server)

        gi9.gi.histories.gi.histories.delete_dataset(
            history_id=history_id, dataset_id=dataset_id, purge=True
        )

    @staticmethod
    def delete_dataset_collection(
         server=None, history_id=None, dataset_collection_id=None
    ):

        a = GiN.sessions.SessionList()
        gi10 = a.get(server=server)

        gi10.gi.histories.gi.histories.delete_dataset_collection(
            history_id=history_id,
            dataset_collection_id=dataset_collection_id,
            # purge=True,
        )


    @staticmethod
    def show_dataset_collection( server=None, dataset_id=None):

        a = GiN.sessions.SessionList()
        gi11 = a.get(server=server)

        show_dataset = gi11.gi.dataset_collections.show_dataset_collection(
            dataset_collection_id=dataset_id
        )


        print(show_dataset)

        # return IPython.display.JSON(show_dataset)
        return show_dataset


    @staticmethod
    def get_histories( server):

        a = GiN.sessions.SessionList()
        gi12 = a.get(server=server)

        histories = gi12.gi.histories.gi.histories.get_histories()

        return IPython.display.JSON(histories)

    @staticmethod
    def download_file_to_jupyter_server(
        server=None,
        data_type="dataset",
        collection_id=None,
        ext="zip",
        file_name=None,
        dir="galaxy_data",
    ):

        a = GiN.sessions.SessionList()
        gi12 = a.get(server=server)

        galaxy_data = os.path.join(os.getcwd(), dir)

        if not os.path.exists(galaxy_data):
            os.mkdir(galaxy_data)

        if data_type == "collection":
            file_path = os.path.join(galaxy_data, file_name) + "." + ext
            gi12.gi.dataset_collections.download_dataset_collection(
                dataset_collection_id=collection_id, file_path=file_path
            )
        else:
            gi12.gi.datasets.download_dataset(
                dataset_id=collection_id, file_path=galaxy_data
            )

    @staticmethod
    def CORS_fallback_upload(
        file_name,
        data, 
        server,
        history_id,
    ):

        temp_dir = os.path.join(os.getcwd(), "temp")

        if not os.path.exists(temp_dir):
            os.mkdir(temp_dir)

        a = GiN.sessions.SessionList()
        gi = a.get(server=server)
        path = os.path.join(temp_dir, file_name)

        with open(path, 'w') as f:
            f.write(data)
       
        f.close()
   
        out = gi.tools.gi.tools.upload_file(path=path, history_id=history_id)

        return IPython.display.JSON(out)

    @staticmethod
    def send_data_to_galaxy_tool(
        server_d=None,
        server_u=None,
        file_name=None,
        dataset_id=None,
        ext="zip",
        dataset_name=None,
        history_id=None,
    ):

        a = GiN.sessions.SessionList()
        gi13 = a.get(server=server_d)
        gi14 = a.get(server=server_u)

        temp_dir = os.path.join(os.getcwd(), "temp1")

        if not os.path.exists(temp_dir):
            os.mkdir(temp_dir)

        for f in os.listdir(temp_dir):
            os.remove(os.path.join(temp_dir, f))

        gi13.gi.datasets.download_dataset(
            dataset_id=dataset_id, file_path=temp_dir, require_ok_state=False
        )

        file_name = glob.glob(temp_dir + "/*.*") 

        # file_name = os.path.join(temp_dir, dataset_name)

        out = gi14.tools.gi.tools.upload_file(path=file_name[0], history_id=history_id)


        return IPython.display.JSON(out)

        # if not os.path.exists(galaxy_data):
        #     os.mkdir(galaxy_data)
        #     gi.gi.datasets.download_dataset(dataset_id=collection_id, file_path=galaxy_data)
    @staticmethod
    def upload_fallback(
        server_u=None,
        file_name=None,
    ):

        gi14 = a.get(server=server_u)
        file_name = glob.glob(temp_dir + "/*.*")
        out = gi14.tools.gi.tools.upload_file(path=file_name[0], history_id=history_id)


        return IPython.display.JSON(out)

    @staticmethod
    def send_data_to_gp_server(file_name, tool_id, dataset_id, server, ext):

        temp_dir = os.path.join(os.getcwd(), "temp1")

        if not os.path.exists(temp_dir):
            os.mkdir(temp_dir)

        for f in os.listdir(temp_dir):
            os.remove(os.path.join(temp_dir, f))

        GalaxyTaskWidget.download_file_to_jupyter_server(
            server=server, collection_id=dataset_id, dir="temp1"
        )

        file_name = glob.glob(os.path.join(temp_dir, "*.*"))
        new_file = os.path.join(
            "/".join(file_name[0].split("/")[: len(file_name[0].split("/")) - 1]),
            "data." + ext,
        )
        os.rename(file_name[0], new_file)
        ####################
        task = GPTask(authwidget.session.sessions[0], tool_id)
        uri = task.server_data.upload_file(
            new_file.split("/")[len(new_file.split("/")) - 1], new_file
        )
        #####################
        return IPython.display.JSON({"uri": uri.uri})

    @staticmethod
    def return_job_status(server=None, job_id=None):

        a = GiN.sessions.SessionList()
        gi15 = a.get(server=server)
        job_state = gi15.gi.jobs.show_job(job_id=job_id, full_details=True)

        return IPython.display.JSON(job_state)

    @staticmethod
    def return_job_state( server=None, job_id=None):

        a = GiN.sessions.SessionList()
        gi16 = a.get(server=server)
        job_state = gi16.gi.jobs.get_state(job_id=job_id)

        return IPython.display.JSON({"job_state": job_state})

class TaskTool(NBTool):

    """Tool wrapper for the authentication widget"""

    def __init__(self, server_name, tool):
        
        NBTool.__init__(self)
        self.origin = server_name
        self.id = tool["id"]
        self.name = tool["name"]
        self.description = tool["description"]
        self.load = lambda: GalaxyTaskWidget(tool)
        



from asyncio import constants
from tkinter import Y
from bioblend.galaxy.objects import GalaxyInstance
from .display import display
from nbtools import UIBuilder, ToolManager, NBTool, EventManager
from threading import Thread
from .sessions import session
# from .shim import login, system_message
from .taskwidget import TaskTool
from .util import DEFAULT_COLOR, DEFAULT_LOGO, GALAXY_SERVERS
from urllib.error import HTTPError
from .Galaxyuibuilder import GalaxyUIBuilder
from ipywidgets import Output
from nbtools.uioutput import UIOutput
from nbtools.event_manager import EventManager
import json


import IPython
import IPython.display

class GalaxyAuthWidget(GalaxyUIBuilder):
    """A widget for authenticating with a Galaxy server"""

    default_color = DEFAULT_COLOR
    default_logo = DEFAULT_LOGO
    new_session=None

    def __init__(self, session=None, **kwargs):
        """Initialize the authentication widget"""

        self.session = session

        GalaxyUIBuilder.__init__(
            self,
            name='login',
            run_label='Login',
            description='Login to Galaxy instance by credentials or API Key',
            # display_header=False,
            color=self.default_color,
            logo=self.default_logo,
            collapsed=False,
            buttons={
                    'Register an Account': '',
            },
            **kwargs
        )


## Registering tools in treads, Login fast but tool loading slow. 

    # def login(self, credentials):

    #     """Login to the Galaxy server"""

    #     tool_list =  {'tools':[]}
    
    #     if credentials['email']:
    #         try:
    #             self.session = GalaxyInstance(credentials['server'], email=credentials['email'], password=credentials['password'])
    #             self.session._notebook_email = credentials['email']
    #         except:
    #             # tool_list['state'] = 'error'
    #             return IPython.display.JSON({'state': 'error'})
    #     else:
    #         try:
    #             self.session = GalaxyInstance(credentials['server'],  api_key=credentials['api_key'], verify=True)
    #             self.session._notebook_email = self.session.gi.users.get_current_user()['email']
    #         except:
    #             # tool_list['state'] = 'error'
    #             return IPython.display.JSON({'state': 'error'})

    #     self.session._notebook_url = credentials['server']
    #     self.session._notebook_password = credentials['password']
    #     self.session._notebook_key = credentials['api_key']

    #     self.register_session()

    #     tool_list['url']  = self.session._notebook_url
    #     tool_list['email'] = self.session._notebook_email    

    #     t = {"id": 'GiN_data_upload_tool',  "description": "Upload data files to galaxy server", "name": "Upload Data", 'origin': self.session._notebook_url, 'inputs': [{'type': 'data_upload'}]}
    #     t = TaskTool('+', t )
    #     ToolManager.instance().register(t)

    #     def register_modules_callback():
    #         for section in self.session.tools.gi.tools.get_tool_panel():
    #             if section["model_class"] == "ToolSection":
    #                 for t in section["elems"]:
    #                     try:
    #                         tool={'id':None, 'description':None, 'name':None}
    #                         if t['model_class'] == 'Tool':
                            
    #                             tool['id'] = t['id']
    #                             tool['description'] = t['description']
    #                             tool['name'] = t['name']+" ("+t['version']+")"
    #                             tool['origin'] = self.session._notebook_url
    #                             tool['email'] = self.session._notebook_email
    #                             tool = TaskTool(tool['origin'], tool)
    #                             ToolManager.instance().register(tool)
    #                             # tool_list['tools'].append(tool)
    #                     except:
    #                         pass
        
    #     registration_thread = Thread(target=register_modules_callback)
    #     registration_thread.start()
                
    #     return IPython.display.JSON({'state':'success'}) 

        

    def login(self, credentials):


        """Login to the Galaxy server"""

        tool_list =  {'tools':[]}
    
        if credentials['email']:
            try:
                self.session = GalaxyInstance(credentials['server'], email=credentials['email'], password=credentials['password'])
                self.session._notebook_email = credentials['email']
            except:
                # tool_list['state'] = 'error'
                return IPython.display.JSON({'state': 'error'})
        else:
            try:
                self.session = GalaxyInstance(credentials['server'],  api_key=credentials['api_key'], verify=True)
                self.session._notebook_email = self.session.gi.users.get_current_user()['email']
            except:
                # tool_list['state'] = 'error'
                return IPython.display.JSON({'state': 'error'})

        self.session._notebook_url = credentials['server']
        self.session._notebook_password = credentials['password']
        self.session._notebook_key = credentials['api_key']

        self.register_session()

        tool_list['url']  = self.session._notebook_url
        tool_list['email'] = self.session._notebook_email        

        for section in self.session.tools.gi.tools.get_tool_panel():
            if section["model_class"] == "ToolSection":
                for t in section["elems"]:
                    tool={'id':None, 'description':None, 'name':None}
                    if t['model_class'] == 'Tool':
                    
                        tool['id'] = t['id']
                        tool['description'] = t['description']
                        tool['name'] = t['name']+" ("+t['version']+")"
                        tool['origin'] = self.session._notebook_url
                        tool['email'] = self.session._notebook_email
                        tool_list['tools'].append(tool)
               
 
        t = {"id": 'GiN_data_upload_tool',  "description": "Upload data files to galaxy server", "name": "Upload Data", 'origin': self.session._notebook_url, 'inputs': [{'type': 'data_upload'}]}
       
        t = TaskTool('+', t )
        ToolManager.instance().register(t)

    
        for i in tool_list['tools']:
            GalaxyAuthWidget().RegisterMod(i)


        return IPython.display.JSON({'state':'success'}) 
   
    def RegisterMod(self, tool):

        # def register_modules_callback():
        #     for i in tools:
        #         # GalaxyAuthWidget().RegisterMod(i)
        tool = TaskTool(tool['origin'], tool)
        ToolManager.instance().register(tool)

        # registration_thread = Thread(target=register_modules_callback)
        # registration_thread.start()
        # return "OK"
        
    
    def has_credentials(self):
        """Test whether the session object is instantiated and whether an email and password have been provided"""
        if type(self.session) is not GalaxyInstance:
            return False  # Test type
        if not self.session._notebook_url:
            return False  # Test server url
        # if not self.session._notebook_email:
        #     print("error", 'email')
        #     return False  # Test email
        if not self.session._notebook_password:
            return False  # Test password
        return True

    def validate_credentials(self, session):
        """Validate the provided credentials"""
        # TODO: Is there a bioblend call to verify the user's login credentials? If so, add it here
        try:
            if session is not None and session.gi.key:
                return True
        except:
            pass
        # except HTTPError:
        #     self.error = 'Invalid username or password. Please try again.'
        #     return False
        # except BaseException as e:
        #     self.error = str(e)
        return False

    def replace_widget(self):
        """Replace the unauthenticated widget with the authenticated widget"""
        # self.form.children[1].value = "" 
       
        # self.form.children[2].value = ""  # Blank password so it doesn't get serialized
        display(GalaxyAuthWidget(session=self.session))  # Display the authenticated widget
        self.close()  # Close the unauthenticated widget

    def register_session(self):
        """Register the validated credentials with the SessionList"""

        self.session = session.register(
            self.session._notebook_url,
            self.session._notebook_email,
            self.session._notebook_password,
            self.session._notebook_key, 
        )

    def system_message(self):
        self.info = "Successfully logged into Galaxy"

    def trigger_login(self):
        """Dispatch a login event after authentication"""
        # Trigger login callbacks of job and task widgets
        # return

        EventManager.instance().dispatch("galaxy.login", self.session)

def server_name(search_url):
    """Search the GALAXY_SERVERS dict for the server with the matching URL"""
    for name, url in GALAXY_SERVERS.items():
        if url == search_url:
            return name
    return search_url

    output = Output()  # Output widget
    
    if message is not None:
        error_msg = message
        name = "Tool origin error"
    else:
        error_msg = f"Cannot find tool: {origin} | {id}"
        name = "Cannot find tool"

    placeholder = UIOutput(
        name= name, 
        error=error_msg,
        color= DEFAULT_COLOR,
        logo= DEFAULT_LOGO,

    )  # Placeholder widget
    output.append_display_data(placeholder)

    # Callback to see if the placeholder needs replaced after a new widget is registered
    def check_registration_callback(data):
        if (
            "origin" in data
            and "id" in data
            and data["origin"] == origin
            and data["id"] == id
        ):
            placeholder.close()
            with output:
                display(self.tool(id=id, origin=origin))

    # Register the callback with the event manager
    EventManager.instance().register(
        "nbtools.register", check_registration_callback
    )
    return output

class AuthenticationTool(NBTool):
    """Tool wrapper for the authentication widget"""

    origin = "+"
    id = "galaxy_authentication"
    name = "Galaxy Login"
    description = "Log into a Galaxy server"
    load = lambda x: GalaxyAuthWidget()


# preventing  "jupyter nbextension install", imports need to be fixed 
try:
    ToolManager.instance().register(AuthenticationTool())
except:
    pass


# from bioblend import galaxy
from bioblend.galaxy.objects import GalaxyInstance
import IPython


class SessionList:
    """
    Keeps a list of all currently registered Galaxy server sessions
    """

    sessions = []

    def register(self, server, email=None, password=None, api_key=None):
        """
        Register a new Galaxy  server session for the provided
        server, email and password. Return the session.
        :param server:
        :param email:
        :param password:
        :return:
        """

        # Create the session
        if (password):
            session = GalaxyInstance(server, email=email, password=password, verify=True) 
            session._notebook_password = password
            
        else:
            session = GalaxyInstance(server,  api_key=api_key,  verify=True)
            
       
        session._notebook_url = server
        session._notebook_email = email
        

        # Validate email if not empty
        valid_email = email is not None

        # Validate that the server is not already registered
        index = self._get_index(server)
        new_server = index == -1

        # Add the new session to the list
        if valid_email and new_server:
            self.sessions.append(session)

        # Replace old session is one exists
        if valid_email and not new_server:
            self.sessions[index] = session

        return session

    def get(self, server):
        """
        Returns a registered GalaxyServer object with a matching Galaxy server url or index
        Returns None if no matching result was found
        :param server:
        :return:
        """

        # Handle indexes
        if isinstance(server, int):
            if server >= len(self.sessions):
                return None
            else:
                return self.sessions[server]

        # Handle server URLs
        index = self._get_index(server)
        if index == -1:
            return None
        else:
            return self.sessions[index]

    def get_servers(self):
        servers = []

        if(self.sessions):
            for session in self.sessions:
                servers.append(session._notebook_url)
            return IPython.display.JSON(servers)
        else:
            return IPython.display.JSON([])

    def clean(self):
        """
        Clear all Galaxy sessions from the sessions list
        :return:
        """
        self.sessions = []

    def _get_index(self, server_url):
        """
        Returns a registered GalaxyServer object with a matching Galaaxy server url
        Returns -1 if no matching result was found
        :param server_url:
        :return:
        """
        for i in range(len(self.sessions)):
            session = self.sessions[i]
            if session._notebook_url == server_url:
                return i
        return -1


"""
Galaxy Sessions Singleton
"""
session = SessionList()


def get_session(index):
    return session.get(index)